from typing import Annotated
from uuid import UUID

from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    HttpUrl,
    computed_field,
)

ADMIN_ROLE = "ADMIN"


class Role(BaseModel):
    name: str
    users: list[str] | None = None


class OrganizationAttributes(BaseModel):
    id: UUID


class Organization(OrganizationAttributes):
    name: str


class UserClaims(BaseModel):
    iss: HttpUrl
    id: UUID = Field(alias="sub")
    sid: str = Field(alias="sid")
    username: str = Field(alias="preferred_username")
    name: str
    given_name: str
    family_name: str
    roles: list[str] = Field(default_factory=list)
    email: str
    email_verified: bool
    scope: Annotated[
        set[str],
        BeforeValidator(lambda v: v.split(" ") if isinstance(v, str) else v),
    ]
    groups: set[str] = Field(default_factory=set)
    organizations: dict[str, OrganizationAttributes] = Field(
        default_factory=dict
    )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def tenant(self) -> str:
        assert self.iss.path is not None
        return self.iss.path.split("/")[-1]

    @computed_field  # type: ignore[prop-decorator]
    @property
    def organization(self) -> Organization | None:
        if self.organizations:
            org_name = next(iter(self.organizations.keys()))
            return Organization.model_validate(
                {
                    "name": org_name,
                    **self.organizations[org_name].model_dump(),
                }
            )
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_admin(self) -> bool:
        return ADMIN_ROLE in self.roles



import pkgutil
from importlib import import_module
from types import ModuleType

from fastloom.db.signals import (
    BaseDocumentSignal,
    Operations,
    SignalsDelete,
    SignalsInsert,
    SignalsUpdate,
)


def init_signals(module: ModuleType):
    if (
        module.__spec__ is None
        or not module.__spec__.submodule_search_locations
    ):
        return
    for i in pkgutil.iter_modules(module.__path__):
        tmp = import_module(f"{module.__name__}.{i.name}")
        if not i.ispkg:
            continue
        init_signals(tmp)


def init_streams(
    models: list[type[BaseDocumentSignal]],
):
    for model_cls in models:
        if (
            model_cls is BaseDocumentSignal
            or model_cls is SignalsInsert
            or model_cls is SignalsUpdate
            or model_cls is SignalsDelete
        ):
            continue
        if issubclass(model_cls, SignalsInsert):
            model_cls.get_publisher(Operations.CREATE)
        if issubclass(model_cls, SignalsUpdate):
            model_cls.get_publisher(Operations.UPDATE)
        if issubclass(model_cls, SignalsDelete):
            model_cls.get_publisher(Operations.DELETE)



import logging
from enum import StrEnum, auto
from typing import TYPE_CHECKING, Any
from uuid import UUID

if TYPE_CHECKING:
    from beanie import (
        Delete,
        Document,
        Insert,
        Replace,
        Save,
        SaveChanges,
        Update,
        after_event,
    )
else:
    try:
        from beanie import (
            Delete,
            Document,
            Insert,
            Replace,
            Save,
            SaveChanges,
            Update,
            after_event,
        )
    except ImportError:
        from pydantic import BaseModel as Delete
        from pydantic import BaseModel as Document
        from pydantic import BaseModel as Insert
        from pydantic import BaseModel as Replace
        from pydantic import BaseModel as Save
        from pydantic import BaseModel as SaveChanges
        from pydantic import BaseModel as Update

        def after_event(*args, **kwargs):
            def decorator(func):
                return func

            return decorator


from pydantic import BaseModel, ConfigDict, PrivateAttr, model_validator

from fastloom.signals.depends import RabbitSubscriber

logger = logging.getLogger(__name__)


class Operations(StrEnum):
    CREATE = auto()
    UPDATE = auto()
    DELETE = auto()


class SignalMessage[T: Document](BaseModel):
    instance: T
    changes: dict[str, Any]
    operation: Operations

    model_config = ConfigDict(ser_json_bytes="base64", val_json_bytes="base64")


class BaseDocumentSignal(Document):
    """
    Assumes that this mixin is used with `BaseDocument` subclasses and
    `BaseDocument` has full state management
    """

    _sent_events: set[tuple[UUID, Operations]] = PrivateAttr(
        default_factory=set
    )
    _PROJECT_NAME: str = ""

    @model_validator(mode="after")
    def validate_state_management(self):
        self.get_settings().use_revision = True
        self.get_settings().use_state_management = True
        self.get_settings().state_management_save_previous = True
        self.check_state_management()  # TODO: unsure which to keep
        return self

    async def _publish(self, message: SignalMessage):
        if self.revision_id is None:
            return
        _event_key = (
            self.revision_id,
            message.operation,
        )
        if _event_key in self._sent_events:
            logger.debug(f"prevented publishing event: {_event_key}")
            return
        logger.debug(f"publishing event: {_event_key}")
        await self.get_publisher(message.operation).publish(
            message,
        )
        self._sent_events.add(_event_key)

    @classmethod
    def get_subscription_topic(cls, operation: Operations):
        return (
            f"{cls._PROJECT_NAME}.{cls.get_collection_name()}."
            f"{operation.value}"
        )

    @classmethod
    def check_state_management(cls):
        if not (
            cls.use_state_management() and cls.state_management_save_previous()
        ):
            raise ValueError(
                f"State management is not enabled for {cls.__name__}"
            )

    @classmethod
    def get_publisher(cls, operation: Operations):
        return RabbitSubscriber.publisher(
            routing_key=cls.get_subscription_topic(operation),
            schema=SignalMessage[cls],  # type: ignore[valid-type]
        )


class SignalsInsert(BaseDocumentSignal):
    @after_event(Insert)
    async def _publish_post_insert(cls):
        await cls.publish_post_insert()

    async def publish_post_insert(self):
        await self._publish(
            SignalMessage(
                instance=self,
                changes=self.get_previous_changes(),
                operation=Operations.CREATE,
            ),
        )


class SignalsUpdate(BaseDocumentSignal):
    @after_event(Replace, SaveChanges, Update, Save)
    async def _publish_post_update(self):
        await self.publish_post_update()

    async def publish_post_update(self):
        await self._publish(
            SignalMessage(
                instance=self,
                changes=self.get_previous_changes(),
                operation=Operations.UPDATE,
            )
        )


class SignalsDelete(BaseDocumentSignal):
    @after_event(Delete)
    async def _publish_post_delete(self):
        await self.publish_post_delete()

    async def publish_post_delete(self):
        await self._publish(
            SignalMessage(
                instance=self,
                changes=self.get_previous_changes(),
                operation=Operations.DELETE,
            )
        )


class SignalsAll(SignalsInsert, SignalsUpdate, SignalsDelete): ...



import secrets


def generate_token(length: int):
    """
    Generates a n-digits numeral token, that used for OTP
    """
    min_value = 10 ** (length - 1)
    max_value = 10**length - 1
    return str(secrets.randbelow(max_value - min_value + 1) + min_value).zfill(
        length
    )


def generate_alphanumeric_token(length: int):
    """
    Generates a n-digits alphanumeric token, that used for OTP
    """
    alphabet = "0123456789abcdefghijklmnopqrstuvwxyz"
    return "".join(secrets.choice(alphabet) for _ in range(length))



from pydantic import (
    BaseModel,
    Field,
    RedisDsn,
)

from fastloom.types import Str


class RedisSettings(BaseModel):
    REDIS_URL: Str[RedisDsn] = Field(
        "redis://localhost:6379/0", validate_default=True
    )



from collections.abc import MutableMapping
from contextlib import suppress
from pathlib import Path
from types import new_class
from typing import TYPE_CHECKING, Annotated, Any, TypeVar

import yaml
from pydantic import BaseModel, RootModel, StringConstraints

from fastloom.auth.depends import (
    JWTAuth,
    OptionalJWTAuth,
)
from fastloom.cache.base import BaseCache, BaseTenantSettingCache
from fastloom.cache.lifehooks import RedisHandler
from fastloom.db.signals import BaseDocumentSignal

if TYPE_CHECKING:
    from aredis_om.model.model import NotFoundError

    from fastloom.db.schemas import BaseTenantSettingsDocument
else:
    try:
        from fastloom.db.schemas import BaseTenantSettingsDocument
    except ImportError:
        BaseTenantSettingsDocument = BaseModel

    try:
        from aredis_om.model.model import NotFoundError
    except ImportError:
        NotFoundError = Exception

from fastloom.cache.settings import RedisSettings
from fastloom.db.settings import MongoSettings
from fastloom.meta import SelfSustaining
from fastloom.settings.base import MonitoringSettings
from fastloom.tenant.depends import (
    BaseGetFrom,
    ContextSource,
    HeaderSource,
    OptionalTokenHeaderSource,
    PathSource,
    TenantDependancySelector,
    TenantNotFound,
    TokenBodySource,
    TokenHeaderSource,
)
from fastloom.tenant.protocols import TenantHostSchema, TenantNameSchema
from fastloom.tenant.utils import SettingCacheSchema

DEFAULT_CONFIG_KEY: str = "default"

TenantName = Annotated[str, StringConstraints(strip_whitespace=True)]
TenantMapping = MutableMapping[TenantName, TenantNameSchema]
TenantMappingWithHosts = MutableMapping[TenantName, TenantHostSchema]


def load_settings[T: BaseModel](
    settings_cls: type[T],
    config_yml_file: Path | None = None,
    defaults_only=False,
) -> MutableMapping[str, T]:
    config_yml_file = config_yml_file or Path.cwd() / "tenants.yaml"
    _loaded_configs: dict[str, Any]
    with config_yml_file.open() as f:
        _loaded_configs = yaml.safe_load(f)

    _default_config: dict[str, Any] = _loaded_configs.pop(
        DEFAULT_CONFIG_KEY, {}
    )
    if defaults_only:
        return {
            DEFAULT_CONFIG_KEY: settings_cls.model_validate(_default_config)
        }  # type: ignore[valid-type]
    return (
        RootModel[dict[str, settings_cls]]  # type: ignore[valid-type]
        .model_validate(
            {
                tenant: _default_config | (config or {})
                for tenant, config in _loaded_configs.items()
            }
        )
        .root
    )


class GetSettingsFrom[V](BaseGetFrom):
    async def _item_getter(self, tenant: str) -> V:
        return await Configs[BaseModel, V].self.get(tenant)  # type: ignore[type-var, misc]


class Configs[T: BaseModel, V: BaseModel](SelfSustaining):
    settings: MutableMapping[str, T]
    general: T
    from_: TenantDependancySelector[T]
    settings_from: GetSettingsFrom[V]
    auth: JWTAuth
    optional_auth: OptionalJWTAuth
    documents_enabled: bool = False
    cache_enabled: bool = False
    service_cls: type[T]
    tenant_cls: type[V]
    # cache
    tenant_schema: SettingCacheSchema[V]

    def __init__(
        self,
        service_cls: type[T],
        tenant_cls: type[V],
    ) -> None:
        if self.self is not None:
            return
        super().__init__()
        self.tenant_cls = tenant_cls
        self.service_cls = service_cls
        self.tenant_schema = SettingCacheSchema(self.tenant_cls)
        self._load_settings_yaml()
        self._load_tenant_yaml()
        self.from_ = self._from_()
        self.settings_from = GetSettingsFrom[V](self.from_)
        self.auth = self._auth()
        self.optional_auth = self._optional_auth()
        self._setup_mongo()
        self._setup_redis()

    def _setup_mongo(self):
        if not issubclass(self.service_cls, MongoSettings):
            return

        self.documents_enabled = True

        if isinstance(self.general, MonitoringSettings):
            BaseDocumentSignal._PROJECT_NAME = self.general.PROJECT_NAME

    def _setup_redis(self):
        if not issubclass(self.service_cls, RedisSettings):
            return

        self.cache_enabled = RedisHandler.enabled

        redis = RedisHandler(self.general).redis
        BaseCache.Meta.database = redis
        BaseTenantSettingCache.Meta.database = redis
        self.tenant_schema.cache.Meta.database = redis

        if isinstance(self.general, MonitoringSettings):
            narrowed_general = self.general
            self.tenant_schema.cache.Meta.model_key_prefix = (
                f"{narrowed_general.PROJECT_NAME}"
            )

    def _load_settings_yaml(self):
        self.settings = load_settings(
            new_class(
                "_SettingsWithTenants",
                (self.service_cls, self.tenant_cls),
            ),
        )
        # ^backward compatibility
        self.general = load_settings(self.service_cls, defaults_only=True)[
            DEFAULT_CONFIG_KEY
        ]

    def _load_tenant_yaml(
        self,
    ):
        self.tenant_schema.config_default = load_settings(
            settings_cls=self.tenant_schema.config,
            defaults_only=True,
        )[DEFAULT_CONFIG_KEY].model_dump()

    def _from_(self) -> TenantDependancySelector[T]:
        return TenantDependancySelector[T](
            settings=self.settings,
            general=self.general,
            source_clses=(
                TokenHeaderSource,
                PathSource,
                HeaderSource,
                ContextSource,
                TokenBodySource,
                OptionalTokenHeaderSource,
            ),
        )

    def _auth(self) -> JWTAuth:
        return JWTAuth(self.general)  # type: ignore[arg-type]

    def _optional_auth(self) -> OptionalJWTAuth:
        return OptionalJWTAuth(self.general)  # type: ignore[arg-type]

    def __getitem__(self, tenant: str):  # farming keks
        return self.get(tenant)

    async def get(self, tenant: str) -> V:
        if self.cache_enabled:
            with suppress(NotFoundError):
                return self.tenant_schema.validate(
                    await self.tenant_schema.cache.get(tenant),
                )
        if self.documents_enabled:
            result = await self.tenant_schema.document.get(tenant)
            if result is not None:
                if self.cache_enabled:
                    await self.tenant_schema.cache.model_validate(
                        result.model_dump()
                    ).save()
                    # ^save in cache for better access time
                return self.tenant_schema.validate(result)
        if tenant in self.settings:
            return self.tenant_schema.model.model_validate(
                self.settings[tenant].model_dump()
            )
        raise TenantNotFound(tenant)

    def __setitem__(self, tenant: str, value: V):  # farming lels
        return self.set(tenant, value)

    async def set(self, tenant: str, value: V):
        # strip defaults before saving
        stripped = self.tenant_schema.strip_defaults(value) | {"id": tenant}
        if self.cache_enabled:
            await self.tenant_schema.cache.model_validate(stripped).save()
        if self.documents_enabled:
            await self.tenant_schema.document.model_validate(stripped).save()


T = TypeVar("T", bound=BaseModel)
ConfigAlias = Configs[T, BaseModel]



from fastloom.auth.settings import IAMSettings
from fastloom.logging.settings import LoggingSettings
from fastloom.settings.base import FastAPISettings, MonitoringSettings


class BaseGeneralSettings(
    IAMSettings, LoggingSettings, MonitoringSettings, FastAPISettings
): ...



import logging

from fastloom.logging.settings import LoggingSettings
from fastloom.logging.utils import (
    ColoredFormatter,
    EndpointFilter,
    QueueFilter,
)


def setup_logging(settings: LoggingSettings):
    logger = logging.getLogger()
    handlers = list(
        filter(lambda h: isinstance(h, logging.StreamHandler), logger.handlers)
    )
    if not handlers:
        stream_handler = logging.StreamHandler()
        logger.addHandler(stream_handler)
        handlers.append(stream_handler)
    for handler in handlers:
        handler.setFormatter(ColoredFormatter())

    app_logger = logging.getLogger(settings.PROJECT_NAME)
    app_logger.setLevel(settings.LOG_LEVEL)

    logging.getLogger("uvicorn.access").addFilter(
        EndpointFilter(settings.LOGGING_EXCLUDED_ENDPOINTS)
    )
    logging.getLogger("faststream.access.rabbit").addFilter(
        QueueFilter(settings.LOGGING_EXCLUDED_QUEUES)
    )



from collections.abc import Callable, Coroutine
from typing import Any

from fastapi import APIRouter, FastAPI, HTTPException
from fastapi.responses import JSONResponse


def init_healthcheck(
    app: FastAPI,
    healthcheck_handlers: list[Callable[[], Coroutine[Any, Any, None]]],
    prefix: str = "",
) -> None:
    router = APIRouter()

    @router.get(f"{prefix}/healthcheck")
    async def healthcheck_endpoint() -> JSONResponse:
        try:
            for handler in healthcheck_handlers:
                await handler()

        except Exception as e:
            raise HTTPException(status_code=503, detail=str(e)) from e

        return JSONResponse(content={"status": "ok"})

    app.include_router(router, tags=["Healthcheck"])



from re import Pattern
from typing import Literal

from fastloom.settings.base import MonitoringSettings


class LoggingSettings(MonitoringSettings):
    LOG_LEVEL: (
        int | Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ) = "INFO"
    LOGGING_EXCLUDED_ENDPOINTS: tuple[Pattern | str, ...] = (
        r"/api/\w+/healthcheck$",
        r"/healthcheck$",
    )
    LOGGING_EXCLUDED_QUEUES: tuple[Pattern | str, ...] = ()



from collections.abc import Awaitable, Callable, Coroutine
from contextlib import asynccontextmanager
from types import ModuleType
from typing import TYPE_CHECKING, Any, Self

from starlette.types import ASGIApp

from fastloom.cache.settings import RedisSettings
from fastloom.db.signals import BaseDocumentSignal
from fastloom.launcher.settings import LauncherSettings
from fastloom.signals.lifehooks import init_signals
from fastloom.signals.settings import RabbitmqSettings
from fastloom.tenant.handler import init_settings_endpoints

if TYPE_CHECKING:
    from beanie import View
    from beanie.odm.documents import Document
    from beanie.odm.union_doc import UnionDoc
else:
    try:
        from beanie import View
        from beanie.odm.documents import Document
        from beanie.odm.union_doc import UnionDoc
    except ImportError:
        from pydantic import BaseModel as Document
        from pydantic import BaseModel as UnionDoc
        from pydantic import BaseModel as View

from fastapi import APIRouter, FastAPI
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import Lifespan

from fastloom.cache.healthcheck import get_healthcheck as cache_hc
from fastloom.db.healthcheck import get_healthcheck as db_hc
from fastloom.db.lifehooks import get_models, init_db
from fastloom.db.settings import MongoSettings
from fastloom.healthcheck.handler import init_healthcheck
from fastloom.i18n.base import CustomI18NException
from fastloom.i18n.handler import i18n_exception_handler
from fastloom.settings.base import FastAPISettings
from fastloom.signals.depends import RabbitSubscriber
from fastloom.signals.healthcheck import (
    get_healthcheck as signal_hc,
)
from fastloom.tenant.settings import ConfigAlias as Configs

Route = tuple[APIRouter, str, str]
SettingsCls = type[BaseModel]
Healthcheck = Callable[[], Coroutine[Any, Any, None]]
ExceptionHandler = Callable[
    [Request, Exception], Response | Awaitable[Response]
]
ExceptionHandlerRegister = tuple[int | type[Exception], ExceptionHandler]


def default_lifespan():
    @asynccontextmanager
    async def _identity_gen(_):
        yield

    return _identity_gen


class App(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    signals_module: ModuleType | None = None
    models_module: ModuleType | None = None
    healthchecks: list[Healthcheck] = Field(default_factory=list)
    additional_instruments: list[Callable] = Field(default_factory=list)
    routes: list[Route] = Field(default_factory=list)
    mounts: list[tuple[str, ASGIApp] | tuple[str, ASGIApp, str]] = Field(
        default_factory=list
    )
    models: list[type[Document] | type[UnionDoc] | type[View]] = Field(
        default_factory=list
    )
    lifespan_fn: Lifespan = Field(default_factory=default_lifespan)
    exception_handlers: list[ExceptionHandlerRegister] = Field(
        default_factory=list
    )
    cache_healthcheck: bool = False

    def load_routes(self, app: FastAPI):
        for router, prefix, name in self.routes:
            app.include_router(
                router,
                prefix=Configs[FastAPISettings].general.API_PREFIX + prefix,  # type: ignore[misc]
                tags=[name],
            )

    def load_mounts(self, app: FastAPI):
        for mount in self.mounts:
            app.mount(*mount)

    async def load(self):
        await self.load_db()
        if self.signals_module is not None:
            init_signals(self.signals_module)

    async def load_db(self):
        if not self.models:
            return
        await init_db(
            database_name=Configs[MongoSettings].general.MONGO_DATABASE,
            models=self.models + [Configs.tenant_schema.document],
            mongo_uri=Configs[MongoSettings].general.MONGO_URI,
        )

    def load_healthchecks(self, app: FastAPI):
        handlers: list[Healthcheck] = [
            *self.healthchecks,
        ]
        if self.cache_healthcheck:
            handlers.append(
                cache_hc(Configs[RedisSettings].general.REDIS_URL)  # type: ignore[misc]
            )
        if self.models:
            handlers.append(db_hc(Configs[MongoSettings].general.MONGO_URI))  # type: ignore[misc]
        if self.signals_module:
            handlers.append(signal_hc(RabbitSubscriber.router))

        init_healthcheck(app=app, healthcheck_handlers=handlers)
        # ^for docker and system healthcheck
        init_healthcheck(
            app=app,
            healthcheck_handlers=handlers,
            prefix=Configs[FastAPISettings].general.API_PREFIX,  # type: ignore[misc]
        )

    def load_system_endpoints(self, app: FastAPI):
        init_settings_endpoints(app=app, configs=Configs)
        if Configs[LauncherSettings].general.SETTINGS_PUBLIC:  # type: ignore[misc]
            init_settings_endpoints(
                app=app,
                configs=Configs,
                prefix=Configs[FastAPISettings].general.API_PREFIX,  # type: ignore[misc]
            )

    def load_exception_handlers(self, app: FastAPI):
        for exc_class_or_status_code, handler in (
            (CustomI18NException, i18n_exception_handler),
            *self.exception_handlers,
        ):
            app.exception_handler(exc_class_or_status_code)(handler)

    @model_validator(mode="after")
    def module_to_models(self) -> Self:
        if self.models_module is not None and not self.models:
            self.models = get_models(self.models_module)
        return self

    @model_validator(mode="after")
    def validate_signals(self) -> Self:
        if self.signals_module is not None and not isinstance(
            Configs[RabbitmqSettings].general,  # type: ignore[misc]
            RabbitmqSettings,
        ):
            raise ValidationError(
                "Settings does not inherit from RabbitmqSettings"
            )
        return self

    @property
    def stream_models(self) -> list[type[BaseDocumentSignal]]:
        return [
            model
            for model in self.models
            if issubclass(model, BaseDocumentSignal)
        ]



import contextlib
import json
import logging
from collections.abc import Callable, Sequence
from enum import Enum
from os import getenv
from typing import TYPE_CHECKING, Any

import logfire
import orjson
from jose.exceptions import JWTError
from jose.jwt import get_unverified_claims
from opentelemetry import metrics, trace
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter,
)
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.trace import Span
from pydantic import AnyHttpUrl, BaseModel
from sentry_sdk import init as sentry_init

from fastloom.cache.settings import RedisSettings
from fastloom.db.settings import MongoSettings
from fastloom.launcher.utils import is_installed
from fastloom.observability.settings import ObservabilitySettings
from fastloom.signals.settings import RabbitmqSettings
from fastloom.tenant.protocols import TenantMonitoringSchema

if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        from fastapi import FastAPI

if not TYPE_CHECKING:
    try:
        from fastapi import FastAPI
    except ImportError:
        from typing import Any as FastAPI


def init_sentry(dsn: AnyHttpUrl | str | None, environment: str):
    if dsn is None:
        return
    if isinstance(dsn, AnyHttpUrl):
        dsn = str(dsn)

    sentry_init(
        dsn=dsn,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        traces_sample_rate=1.0,
        enable_tracing=True,
        profiles_sample_rate=1.0,
        environment=environment,
        send_default_pii=True,
    )


def get_metrics_reader() -> PeriodicExportingMetricReader:
    return PeriodicExportingMetricReader(OTLPMetricExporter())


def _get_authorization_header(scopes: dict[str, Any]) -> str | None:
    headers: dict[str, Any] = {
        key.lower().decode("latin-1"): value.decode("latin-1")
        for key, value in scopes["headers"]
    }
    if "authorization" not in headers:
        return None
    return headers["authorization"]


def _set_user_attributes_to_span(span: Span, token: str):
    try:
        payload: dict[str, Any] = get_unverified_claims(token)
        span.set_attribute("username", payload["name"])
        span.set_attribute("user_id", payload["sub"])
        span.set_attribute("tenant", payload["owner"])
    except (JWTError, KeyError):
        return


def instrument_fastapi(app: FastAPI):
    from fastapi.responses import PlainTextResponse
    from fastapi.security.utils import get_authorization_scheme_param
    from starlette.exceptions import HTTPException as StarletteHTTPException

    def _server_request_hook(span: Span, scope: dict):
        if (
            span
            and span.is_recording()
            and (auth_header := _get_authorization_header(scope))
        ):
            scheme, param = get_authorization_scheme_param(auth_header)
            if scheme.lower() != "bearer":
                return
            _set_user_attributes_to_span(span, param)

    def _client_response_hook(span: Span, scope: dict, message: dict):
        if span and span.is_recording():
            ...

    logfire.instrument_fastapi(
        app,
        server_request_hook=_server_request_hook,
        client_response_hook=_client_response_hook,
        meter_provider=metrics.get_meter_provider(),
    )

    @app.exception_handler(StarletteHTTPException)
    async def http_exception_handler(request, exc: StarletteHTTPException):
        current_span = trace.get_current_span()
        if current_span is not None and current_span.is_recording():
            current_span.set_attributes(
                {
                    "http.status_text": str(exc.detail),
                    "otel.status_description": (
                        f"{exc.status_code} / {str(exc.detail)}"
                    ),
                    "otel.status_code": "ERROR",
                }
            )
        return PlainTextResponse(
            json.dumps({"detail": str(exc.detail)}),
            status_code=exc.status_code,
        )


def instrument_logging(settings):
    class AttributedLogfireLoggingHandler(logfire.LogfireLoggingHandler):
        def fill_attributes(self, record: logging.LogRecord):
            record.SERVICE_NAME = settings.PROJECT_NAME
            record.HOST_NAME = settings.ENVIRONMENT
            return super().fill_attributes(record)

    logger = logging.getLogger()
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logfire_handler = AttributedLogfireLoggingHandler()
    logfire_handler.setFormatter(formatter)
    logger.addHandler(logfire_handler)


def instrument_metrics():
    logfire.instrument_system_metrics(base="basic")


def instrument_httpx():
    logfire.instrument_httpx(tracer_provider=trace.get_tracer_provider())


def instrument_requests():
    logfire.instrument_requests(tracer_provider=trace.get_tracer_provider())


def instrument_redis():
    logfire.instrument_redis(
        capture_statement=True, tracer_provider=trace.get_tracer_provider()
    )


def instrument_celery():
    logfire.instrument_celery(
        tracer_provider=trace.get_tracer_provider(),
        meter_provider=metrics.get_meter_provider(),
    )


def instrument_confluent_kafka():
    from opentelemetry.instrumentation.confluent_kafka import (
        ConfluentKafkaInstrumentor,
    )

    ConfluentKafkaInstrumentor().instrument(
        tracer_provider=trace.get_tracer_provider()
    )


def instrument_rabbit():  # TODO: check if logfire picks this up
    from opentelemetry.instrumentation.aio_pika import AioPikaInstrumentor

    AioPikaInstrumentor().instrument(
        tracer_provider=trace.get_tracer_provider()
    )


def instrument_mongodb():
    from bson import DBRef, Decimal128, ObjectId
    from pymongo import monitoring

    def parse_mongo_types(obj: object):
        if isinstance(obj, Decimal128):
            return str(obj.to_decimal())
        if ObjectId.is_valid(obj):
            return str(obj)
        if isinstance(obj, DBRef):
            return str(obj)
        raise TypeError()

    def _response_hook(span: Span, event: monitoring.CommandSucceededEvent):
        if span and span.is_recording():
            span.set_attribute(
                "db.mongodb.server_reply",
                orjson.dumps(
                    event.reply,
                    default=parse_mongo_types,
                    option=orjson.OPT_NAIVE_UTC,
                ),
            )

    logfire.instrument_pymongo(
        tracer_provider=trace.get_tracer_provider(),
        capture_statement=True,
        response_hook=_response_hook,
    )


def instrument_openai(client: Any | None = None):
    from openai import AsyncOpenAI, OpenAI

    if client is not None and not isinstance(client, (OpenAI, AsyncOpenAI)):
        raise ValueError("client must be an instance of OpenAI or AsyncOpenAI")

    logfire.instrument_openai(client)


def instrument_pydantic_ai():
    logfire.instrument_pydantic_ai()


class Instruments(Enum):
    REDIS = instrument_redis
    CELERY = instrument_celery
    RABBIT = instrument_rabbit
    HTTPX = instrument_httpx
    REQUESTS = instrument_requests
    METRICS = instrument_metrics
    MONGODB = instrument_mongodb
    OPENAI = instrument_openai
    PYDANTIC_AI = instrument_pydantic_ai


def instrument_otel(
    settings: TenantMonitoringSchema,
    app: FastAPI | None = None,
    only: Sequence[Instruments] = (),
):
    logfire.configure(
        send_to_logfire="if-token-present",
        service_name=settings.PROJECT_NAME,
        environment=settings.ENVIRONMENT,
        console=False,
        metrics=logfire.MetricsOptions(
            additional_readers=[get_metrics_reader()]
        )
        if getenv("OTEL_EXPORTER_OTLP_ENDPOINT") is not None
        else None,
    )

    instrument_logging(settings)
    if app:
        instrument_fastapi(app)
    for item in only:
        instrument: Instruments
        args: Sequence[Any] | None = None
        if isinstance(item, Sequence):
            instrument, args = item
        else:
            instrument = item
        func: Callable = (
            instrument if callable(instrument) else instrument.value
        )
        func(*args) if args is not None else func()


def infer_instruments[T: BaseModel](settings: T) -> list[Instruments]:
    instruments = []
    if is_installed("httpx"):
        instruments.append(Instruments.HTTPX)
    if isinstance(settings, RedisSettings):
        instruments.append(Instruments.REDIS)
    if isinstance(settings, RabbitmqSettings):
        instruments.append(Instruments.RABBIT)
    if isinstance(settings, MongoSettings):
        instruments.append(Instruments.MONGODB)
    if isinstance(settings, ObservabilitySettings) and settings.METRICS:
        instruments.append(Instruments.METRICS)
    if is_installed("pydantic_ai"):
        instruments.append(Instruments.PYDANTIC_AI)
    return instruments


class InitMonitoring:
    def __init__(
        self,
        settings: ObservabilitySettings,
        instruments: Sequence[Instruments] = (),
    ):
        self.settings = settings
        self.instruments = instruments

    def __enter__(self):
        if int(self.settings.SENTRY_ENABLED):
            init_sentry(self.settings.SENTRY_DSN, self.settings.ENVIRONMENT)

        if int(self.settings.OTEL_ENABLED):
            instrument_otel(
                self.settings,
                only=self.instruments + infer_instruments(self.settings),
            )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb): ...

    def instrument(self, app: FastAPI):
        if app is not None and int(self.settings.OTEL_ENABLED):
            instrument_fastapi(app)



from datetime import datetime
from typing import TYPE_CHECKING, Annotated

if TYPE_CHECKING:
    from beanie import (
        Document,
        Indexed,
        Insert,
        Replace,
        SaveChanges,
        Update,
        before_event,
    )
else:
    try:
        from beanie import (
            Document,
            Indexed,
            Insert,
            Replace,
            SaveChanges,
            Update,
            before_event,
        )
    except ImportError:
        from pydantic import BaseModel as Document
        from pydantic import BaseModel as Insert
        from pydantic import BaseModel as Replace
        from pydantic import BaseModel as SaveChanges
        from pydantic import BaseModel as Update

        def Indexed() -> type:
            return str

        def before_event(*args, **kwargs):
            def decorator(func):
                return func

            return decorator


from pydantic import BaseModel, Field, computed_field, field_validator

from fastloom.date import utcnow


class CreatedAtSchema(BaseModel):
    created_at: datetime = Field(default_factory=utcnow)


class CreatedUpdatedAtSchema(CreatedAtSchema):
    """
    ONLY use this mixin in `beanie.Document` models since it uses
    @before_event decorator

    NOTE: `updated_at` doesn't get updated when `update_many` is called
    """

    updated_at: datetime | None = Field(default_factory=utcnow)
    # TODO ^ it shouldn't ideally be None, but some models used to save null
    # so first we have to make sure we cleared db from all such instances

    @before_event(Insert, Replace, SaveChanges, Update)
    async def update_updated_at(self):
        self.updated_at = utcnow()


class BasePaginationQuery(BaseModel):
    offset: int | None = Field(None, ge=0)
    limit: int | None = Field(None, ge=0)

    @field_validator("limit", mode="after")
    @classmethod
    def convert_zero_limit(cls, v: int | None) -> int | None:
        return v or None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def skip(self) -> int | None:
        if self.limit and self.offset is not None:
            return self.limit * self.offset
        return None


class PaginatedResponse[T](BaseModel):
    data: list[T] = Field(default_factory=list)
    count: int = Field(default=0, ge=0)


class BaseTenantSettingsDocument(CreatedUpdatedAtSchema, Document):
    id: Annotated[str, Indexed()]  # type: ignore[assignment]

    class Settings:
        name = "settings"



import logging
import re
from re import Pattern

import click


class ColoredFormatter(logging.Formatter):
    @staticmethod
    def get_colored_levelname(level_name, levelno):
        level_colors = {
            logging.DEBUG: lambda text: click.style(text, fg="bright_blue"),
            logging.INFO: lambda text: click.style(text, fg="green"),
            logging.WARNING: lambda text: click.style(text, fg="yellow"),
            logging.ERROR: lambda text: click.style(text, fg="red"),
            logging.CRITICAL: lambda text: click.style(text, fg="bright_red"),
        }
        return level_colors.get(levelno, lambda text: text)(level_name)

    def format(self, record):
        formatted_message = super().format(record)
        levelname = self.get_colored_levelname(
            record.levelname, record.levelno
        )
        name = click.style(record.name, fg="magenta")
        message = click.style(formatted_message, fg="bright_cyan")

        colorized_message = f"{levelname}: \t  {name} {message}"
        return colorized_message


class EndpointFilter(logging.Filter):
    def __init__(
        self, excluded_endpoints: tuple[Pattern | str, ...] = ()
    ) -> None:
        super().__init__()
        self.excluded_endpoints = excluded_endpoints

    def filter(self, record: logging.LogRecord) -> bool:
        if not record.args or len(record.args) < 3:
            return True
        endpoint: str = record.args[2]  # type: ignore[assignment, index]
        return not any(
            re.match(exp, endpoint) for exp in self.excluded_endpoints
        )


class QueueFilter(logging.Filter):
    def __init__(self, excluded_queues: tuple[Pattern | str, ...] = ()):
        super().__init__()
        self.excluded_queues = excluded_queues

    def filter(self, record: logging.LogRecord) -> bool:
        queue: str = getattr(record, "queue", "")
        if record.levelno >= logging.ERROR:
            return True
        return not any(re.match(exp, queue) for exp in self.excluded_queues)



from abc import abstractmethod
from collections.abc import Callable, MutableMapping
from itertools import chain
from json import JSONDecodeError
from typing import Annotated

from aredis_om.model.model import NotFoundError  # type: ignore[import-untyped]
from fastapi import Depends, Header, HTTPException, Path, Request
from pydantic import StringConstraints

from fastloom.auth.depends import JWTAuth, OptionalJWTAuth
from fastloom.auth.schemas import UserClaims
from fastloom.cache.base import HostTenantMapping
from fastloom.cache.lifehooks import RedisHandler
from fastloom.tenant import Tenant
from fastloom.tenant.protocols import TenantHostSchema, TenantNameSchema

TenantName = Annotated[str, StringConstraints(strip_whitespace=True)]
TenantMapping = MutableMapping[TenantName, TenantNameSchema]
TenantMappingWithHosts = MutableMapping[TenantName, TenantHostSchema]
type SettingsMapping[K] = MutableMapping[TenantName, K]
type SettingsMappingGetter[K] = Callable[[], SettingsMapping[K]]


class TenantNotFound(Exception):
    def __init__(self, tenant: str):
        self.tenant = tenant

    def __str__(self):
        return f"Tenant {self.tenant} not found in settings"


class BaseTenantSource[K]:
    settings: MutableMapping[TenantName, K]
    general: K

    def __init__(
        self, settings: MutableMapping[TenantName, K], general: K
    ) -> None:
        self.settings = settings
        self.general = general

    @abstractmethod
    async def _dep(self, *args, **kwargs) -> str | None:
        pass

    def get_dep(self) -> Callable[..., str | None]:
        def _inner(
            tenant: Annotated[str | None, Depends(self._dep)],
        ) -> str | None:
            if tenant is None:
                return None
            if tenant not in self.settings:
                raise TenantNotFound(tenant)
            return tenant

        return _inner


class HeaderSource(BaseTenantSource[TenantHostSchema]):
    async def _dep(
        self, x_forwarded_host: Annotated[str, Header(include_in_schema=False)]
    ) -> str | None:
        if RedisHandler.enabled:
            try:
                tenant = (await HostTenantMapping.get(x_forwarded_host)).tenant
            except NotFoundError:
                tenant = self.hosts[x_forwarded_host]
        else:
            tenant = self.hosts[x_forwarded_host]
        Tenant.set(tenant)
        return tenant

    @property
    def hosts(self) -> dict[str, str]:
        return dict(
            chain(
                *(
                    tuple(
                        (url.host, tenant.name)
                        for url in (
                            tenant.website_url
                            if isinstance(tenant.website_url, list)
                            else [tenant.website_url]
                        )
                        if url.host
                    )
                    for tenant in self.settings.values()
                )
            )
        )


class PathSource(BaseTenantSource):
    async def _dep(self, tenant: Annotated[str, Path()]) -> str:
        Tenant.set(tenant)
        return tenant


class TokenBodySource(BaseTenantSource):
    async def _dep(self, req: Request) -> str:
        try:
            if "token" not in (req_json := await req.json()):
                raise HTTPException(
                    status_code=400, detail="Token not found in request body."
                )
        except JSONDecodeError as er:
            raise HTTPException(
                status_code=400, detail="Request body is not JSON decodable."
            ) from er

        tenant = self.auth._parse_token(req_json["token"]).tenant
        Tenant.set(tenant)
        return tenant

    @property
    def auth(self) -> JWTAuth:
        return JWTAuth(self.general)


class OptionalTokenHeaderSource(BaseTenantSource):
    def get_dep(self) -> Callable[..., str | None]:
        def _inner(
            claims: Annotated[
                UserClaims | None, Depends(self.auth.get_claims)
            ],
        ) -> str | None:
            if claims is None:
                return None
            return self._get_tenant_from_claims(claims)

        return _inner

    def _get_tenant_from_claims(self, claims: UserClaims) -> str:
        Tenant.set(claims.tenant)
        return claims.tenant

    @property
    def auth(self) -> OptionalJWTAuth:
        return OptionalJWTAuth(self.general)


class TokenHeaderSource(OptionalTokenHeaderSource):
    def get_dep(self) -> Callable[..., str]:
        def _inner(
            claims: Annotated[UserClaims, Depends(self.auth.get_claims)],
        ) -> str:
            return self._get_tenant_from_claims(claims)

        return _inner


try:
    from faststream import Depends as StreamDepends
    from faststream.rabbit.fastapi import RabbitMessage

    class ContextSource(BaseTenantSource):
        async def _dep(self, tenant: Annotated[str, RabbitMessage]):
            return tenant

    def get_dep(self) -> Callable[..., str | None]:
        def _inner(
            tenant: Annotated[str, StreamDepends(self._dep)],
        ) -> str | None:
            Tenant.set(tenant)
            return tenant

        return _inner

except ImportError:
    pass


class TenantDependancySelector[K]:
    settings: MutableMapping[TenantName, K]
    general: K
    source_clses: tuple[type[BaseTenantSource], ...]

    def __init__(
        self,
        settings: MutableMapping[TenantName, K],
        general: K,
        source_clses: tuple[type[BaseTenantSource], ...],
    ) -> None:
        self.settings = settings
        self.general = general
        self.source_clses = source_clses

    def __getitem__(
        self, source_cls: type[BaseTenantSource]
    ) -> Callable[..., str | None]:
        return self.sources[source_cls.__name__].get_dep()

    @property
    def sources(self) -> dict[TenantName, BaseTenantSource]:
        return {
            source_cls.__name__: source_cls(self.settings, self.general)
            for source_cls in self.source_clses
        }


class BaseGetFrom[K]:
    dep_selector: TenantDependancySelector

    def __init__(self, dep_selector: TenantDependancySelector) -> None:
        self.dep_selector = dep_selector

    @abstractmethod
    async def _item_getter(self, tenant: str):
        raise NotImplementedError("Must implement _item_getter method")

    def __getitem__(self, source_cls: type[BaseTenantSource]):
        async def _inner(
            tenant: Annotated[str, Depends(self.dep_selector[source_cls])],
        ) -> K:
            return await self._item_getter(tenant)

        return _inner



from collections.abc import Awaitable, Callable
from functools import wraps
from os import getpid

from fastloom.cache.lifehooks import RedisHandler
from fastloom.settings.base import ProjectSettings
from fastloom.tenant.settings import ConfigAlias as Configs


class RedisGuardGate:
    """
    - *context manager*:
    ```
    async with RedisGuardGate("boostrap", ttl=30, grace=10) as acquired:
        if acquired:
            await lifespan_init()
    ```
    - *decorator*:
    ```
    @RedisGuardGate("boostrap", ttl=30)
    async def lifespan_init():
        ...
    ```
    """

    key: str
    ttl: int
    grace: int
    _acquired: bool = False

    def __init__(self, key: str, ttl: int = 60, grace: int = 0):
        self.key = (
            f"{Configs[ProjectSettings].general.PROJECT_NAME}:{key}:leader"  # type: ignore[misc]
        )
        self.ttl = ttl
        self.grace = grace

    def __call__[T, **P](self, func: Callable[P, Awaitable[T]]):
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | None:
            async with self as acquired:
                if acquired:
                    return await func(*args, **kwargs)
                return None

        return wrapper

    async def _acquire(self):
        acquired = await RedisHandler.redis.set(
            self.key,
            str(getpid()),
            nx=True,
            ex=self.ttl,
        )
        return acquired is not None

    async def _release(self):
        await RedisHandler.redis.expire(self.key, self.grace)

    async def __aenter__(self) -> bool:
        self._acquired = await self._acquire()
        return self._acquired

    async def __aexit__(self, exc_type, exc, tb):
        if self._acquired:
            await self._release()



import logging
from typing import Any

from aio_pika import Message
from faststream import ExceptionMiddleware
from faststream.rabbit import (
    ExchangeType,
    RabbitExchange,
    RabbitMessage,
    RabbitQueue,
)
from faststream.rabbit.fastapi import RabbitRouter
from faststream.rabbit.publisher.usecase import RabbitPublisher
from faststream.rabbit.schemas.queue import ClassicQueueArgs
from opentelemetry import trace

from fastloom.meta import SelfSustaining
from fastloom.settings.base import MonitoringSettings
from fastloom.signals.middlewares import RabbitPayloadTelemetryMiddleware
from fastloom.signals.settings import RabbitmqSettings

logger = logging.getLogger(__name__)


def get_rabbit_router(name: str, settings: RabbitmqSettings) -> RabbitRouter:
    return RabbitRouter(
        settings.RABBIT_URI,
        schema_url=f"{name}/asyncapi",
        middlewares=(
            RabbitPayloadTelemetryMiddleware(
                tracer_provider=trace.get_tracer_provider()
            ),
        ),
    )


class RabbitSubscriptable(MonitoringSettings, RabbitmqSettings): ...


class RabbitSubscriber(SelfSustaining):
    """A class to encapsulate the common logic for RabbitMQ subscribers"""

    router: RabbitRouter
    exchange: RabbitExchange
    exc_middleware: ExceptionMiddleware
    _settings: RabbitSubscriptable
    _base_delay: int
    _max_delay: int
    _queue_prefix: str

    def __init__(
        self,
        settings: RabbitSubscriptable,
        base_delay: int = 5,
        max_delay: int = 3600 * 24,
        exceptions: list[type[Exception]] | None = None,
    ):
        """
        :param settings: settings object with
        RABBIT_URI, ENVIRONMENT, PROJECT_NAME
        :param base_delay: base delay for retrying messages
        :param max_delay: max delay for retrying messages
        :param exceptions: list of exceptions to retry

        creates a RabbitMQ router, exchange, and exception middleware \n
        example usage:
            ```
            rabbit_subscriber = RabbitSubscriber(settings)
            app.include_router(rabbit_subscriber.router)

            @rabbit_subscriber.subscriber("routing_key", with_backoff=True)
            async def handler(message: StreamMessage[IncomingMessage]): ...
            ```
        """
        super().__init__()
        self._settings = settings
        if exceptions is None:
            exceptions = [Exception]
        self.router = get_rabbit_router(
            self._settings.API_PREFIX, self._settings
        )
        self.exchange = RabbitExchange(
            name="amq.topic", type=ExchangeType.TOPIC, durable=True
        )
        self.exc_middleware = ExceptionMiddleware(
            handlers={exc: self._exc_handler for exc in exceptions}
        )
        self.router.broker.add_middleware(self.exc_middleware)
        self._base_delay = base_delay
        self._max_delay = max_delay
        self._queue_prefix = (
            f"{self._settings.ENVIRONMENT}_{self._settings.PROJECT_NAME}"
        )

    @classmethod
    def _get_queue_name(cls, name: str) -> str:
        return f"{cls._queue_prefix}.{name}"

    @classmethod
    def _dlx_suffix(
        cls,
    ) -> str:
        return f".{cls._settings.PROJECT_NAME}"

    @classmethod
    def _get_dlx_name(
        cls,
        routing_key: str,
    ) -> str:
        return f"{cls._get_queue_name(routing_key)}{cls._dlx_suffix()}"

    @classmethod
    def _sanitize_routing_key(cls, routing_key: str) -> str:
        return routing_key.replace("*", "__all__")

    @classmethod
    def _get_queue(cls, name: str) -> RabbitQueue:
        """
        :param name: name of the queue
        :return: RabbitQueue
        """
        return RabbitQueue(
            name=cls._get_queue_name(cls._sanitize_routing_key(name)),
            routing_key=name,
            durable=True,
        )

    @classmethod
    async def _get_dlx_queue(
        cls,
        routing_key: str,
        delay: int,
    ) -> RabbitQueue:
        """
        :param routing_key: routing key for the queue
        :param delay: delay in seconds
        :return: RabbitQueue

        creates a dead letter queue with specified delay
        and binds it to the exchange
        """

        queue = RabbitQueue(
            name=f"{cls._get_dlx_name(cls._sanitize_routing_key(routing_key))}.{delay}",
            routing_key=f"{routing_key}{cls._dlx_suffix()}.{delay}",
            durable=True,
            arguments=ClassicQueueArgs(
                {
                    "x-dead-letter-exchange": cls.exchange.name,
                    "x-dead-letter-routing-key": (
                        f"{routing_key}{cls._dlx_suffix()}"
                    ),
                    "x-message-ttl": delay * 1000,
                    # "x-expires": delay * 2000,
                }
            ),
        )

        robust_queue = await cls.router.broker.declare_queue(
            queue=queue,
        )
        dlx = await cls.router.broker.declare_exchange(cls.exchange)
        await robust_queue.bind(
            dlx,
            routing_key=f"{routing_key}{cls._dlx_suffix()}.{delay}",
        )

        return queue

    @classmethod
    async def _exc_handler(
        cls,
        exc: Exception,
        message: RabbitMessage,
    ):
        message.headers["x-delivery-count"] = (
            message.headers.get("x-delivery-count", 0) + 1
        )
        if message.raw_message.routing_key is None:
            raise exc
        if (routing_key := message.raw_message.routing_key).endswith(
            cls._dlx_suffix()
        ) and message.headers["x-delivery-count"] > 1:
            routing_key = routing_key[: -len(cls._dlx_suffix())]

        queue = await cls._get_dlx_queue(
            routing_key,
            min(
                cls._base_delay
                * 2 ** (message.headers["x-delivery-count"] - 1),
                cls._max_delay,
            ),
        )

        await cls.router.broker.publish(
            Message(body=message.body, headers=message.headers),
            queue=queue,
            exchange=cls.exchange,
            persist=True,
        )
        # re-raise for observability in sentry/otel
        raise exc

    @classmethod
    def _get_subscriber(cls, routing_key: str, **kwargs):
        return cls.router.subscriber(
            queue=cls._get_queue(routing_key),
            exchange=cls.exchange,
            **kwargs,
        )

    @classmethod
    def subscriber(
        cls,
        routing_key: str,
        retry_backoff: bool = False,
        **kwargs,
    ):
        """
        :param routing_key: routing key for the queue
        :param retry_backoff: whether to retry with backoff
        :param kwargs: additional faststream subscriber arguments
        :return: custom decorator for the subscriber
        """

        def _inner(func):
            decorators = [cls._get_subscriber(routing_key, **kwargs)]
            if retry_backoff:
                decorators.append(
                    cls._get_subscriber(
                        f"{routing_key}{cls._dlx_suffix()}",
                        **kwargs,
                    )
                )
            for decorator in decorators:
                func = decorator(func)

            return func

        return _inner

    @classmethod
    def publisher(
        cls,
        routing_key: str,
        schema: Any | None = None,
        **kwargs,
    ):
        """
        :param routing_key: routing key
        :param schema : pydantic schema
        :param kwargs: additional faststream subscriber arguments
        :return: persistent publisher
        """
        return cls.router.publisher(
            routing_key=routing_key,
            exchange=cls.exchange,
            schema=schema,
            persist=True,
            **kwargs,
        )

    @classmethod
    def multi_subscriber(
        cls,
        routing_keys: list[str],
        retry_backoff: bool = False,
        **kwargs,
    ):
        """
        :param routing_keys: list of routing keys for the subscribers
        :param retry_backoff: whether to retry with backoff
        :param kwargs: additional faststream subscriber arguments
        """

        def _inner(func):
            for routing_key in routing_keys:
                func = cls.subscriber(
                    routing_key,
                    retry_backoff=retry_backoff,
                    **kwargs,
                )(func)
            return func

        return _inner

    @classmethod
    def multi_publisher(
        cls,
        routing_keys: dict[str, str],
        schema: type | None = None,
        **kwargs,
    ) -> dict[str, RabbitPublisher]:
        """
        :param routing_keys: list of routing keys for the publishers
        :param schema: publish schema for the publishers
        :param kwargs: additional arguments for the faststream publishers
        """
        return {
            key: cls.router.publisher(
                exchange=cls.exchange,
                routing_key=routing_key,
                schema=schema,
                persist=True,
                **kwargs,
            )
            for key, routing_key in routing_keys.items()
        }



from enum import StrEnum

from beanie import Document
from pydantic import model_validator
from pymongo import IndexModel

from fastloom.db.schemas import CreatedUpdatedAtSchema
from fastloom.file.schema import (
    BaseFile,
    FileMetaData,
    RequiredMediaPath,
)
from fastloom.tenant.schemas import TenantMixin


class FileObject(Document, TenantMixin):
    class Settings:
        name = "files"
        indexes = [IndexModel(["name", "tenant"], unique=True)]

    name: str
    usage: StrEnum
    metadata: FileMetaData
    content_type: str
    path: RequiredMediaPath


class FileReference(Document, TenantMixin, BaseFile, CreatedUpdatedAtSchema):
    class Settings:
        name = "file_references"
        indexes = [
            IndexModel(
                ["created_at"],
                expireAfterSeconds=60 * 60 * 24 * 7,  # 7 days
                partialFilterExpression=dict(matched=False),
                name="expire_after_7_days",
            ),
        ]

    usage: StrEnum

    @model_validator(mode="after")
    def validate_content_length_and_matched(self):
        if self.matched and self.content_length is None:
            self.content_length = -1
        return self



from collections.abc import Callable, Coroutine
from functools import partial
from typing import Any


class RedisConnectionError(Exception): ...


async def check_redis_connection(redis_url: str) -> None:
    from redis.asyncio import Redis

    try:
        client: Redis = Redis.from_url(redis_url)
        await client.ping()
    except Exception as er:
        raise RedisConnectionError(f"Redis connection error: {er}") from er


def get_healthcheck(
    redis_url: str,
) -> Callable[[], Coroutine[Any, Any, None]]:
    return partial(check_redis_connection, redis_url=redis_url)



import json
from collections.abc import Generator
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    model_validator,
)

from fastloom.types import NoneField


def _file_to_path(
    file: "str | dict[str, Any] | RequiredMediaPath | OptionalFileField",
) -> "MediaPath":
    if file is None:
        return None

    if isinstance(file, str):
        return RequiredMediaPath(file)

    if isinstance(file, Path):
        return file

    if isinstance(file, dict):
        return _file_to_path(file.get("path"))

    return file.path if file.path else file.name


# TODO remove when FE is compatible
MediaPath = Annotated[
    Path | str | None,
    BeforeValidator(_file_to_path),
    PlainSerializer(lambda p: str(p) if p else None, return_type=str | None),
]
RequiredMediaPath = Annotated[Path, PlainSerializer(str, return_type=str)]


class FileMetaData(BaseModel):
    user_id: str
    username: str


class Filename(BaseModel):
    name: str


class FileContentData(BaseModel):
    content_type: str | None = None
    content_length: int | None = None


class FileMessage(Filename, FileContentData):
    name: str
    usage: StrEnum
    metadata: FileMetaData
    path: RequiredMediaPath
    tenant: str = Field(validation_alias="bucket")


FileIn = Annotated[Filename | None, Field(None)]


class BaseFile(Filename, FileContentData):
    path: MediaPath = None
    matched: bool

    @model_validator(mode="after")
    def validate_content_length_and_matched(self):
        assert (
            (self.path is None)
            == (self.content_type is None)
            == (self.content_length is None)
            == (not self.matched)
        ), json.dumps(
            {
                "path": str(self.path),
                "content_type": self.content_type,
                "content_length": self.content_length,
                "matched": self.matched,
            }
        )
        return self


class UnmatchedFile(BaseFile):
    path: NoneField = None
    content_type: NoneField = None
    content_length: NoneField = None
    matched: Literal[False] = False


class MatchedFile(BaseFile):
    path: RequiredMediaPath = Field(default_factory=Path)
    content_type: str = Field(default_factory=str)
    content_length: int = -1
    matched: Literal[True] = True


FileField = Annotated[
    MatchedFile | UnmatchedFile, Field(union_mode="left_to_right")
]

OptionalFileField = Annotated[FileField | None, Field(None)]

FileFind: TypeAlias = Generator[FileField, Any, None]



from fastapi.openapi.models import OAuthFlow, OAuthFlows
from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
    computed_field,
)

from fastloom.types import Str


class OAuth2MergedScheme(OAuthFlow):
    authorizationUrl: Str[HttpUrl] | None = None
    tokenUrl: Str[HttpUrl] | None = None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def flows(self) -> OAuthFlows:
        if self.authorizationUrl is None and self.tokenUrl is None:
            return OAuthFlows()
        return OAuthFlows.model_validate(
            dict(
                authorizationCode=self.model_dump(
                    exclude_computed_fields=True
                ),
            )
        )
        # ^ implicit & ROPC are deprecated in OAUTH2.1

    @computed_field  # type: ignore[prop-decorator]
    @property
    def oauth2_enabled(self) -> bool:
        return self.authorizationUrl is not None and self.tokenUrl is not None


class OIDCCScheme(BaseModel):
    OIDC_URL: Str[HttpUrl] | None = None

    @computed_field  # type: ignore[misc]
    @property
    def oidc_enabled(self) -> bool:
        return self.OIDC_URL is not None


class IntrospectionResponse(BaseModel):
    active: bool


class IAMSettings(OAuth2MergedScheme, OIDCCScheme):
    INTROSPECT: bool = False
    ACL: bool = False
    IAM_SIDECAR_URL: Str[HttpUrl] | None = Field(None, validate_default=True)



from gettext import gettext as _
from typing import Any

from fastapi import APIRouter, FastAPI, HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

from fastloom.i18n.base import DoesNotExist
from fastloom.launcher.utils import reload_app
from fastloom.tenant.depends import TenantNotFound
from fastloom.tenant.settings import Configs


def init_settings_endpoints(
    app: FastAPI,
    configs: type[Configs[Any, BaseModel]],
    prefix: str = "",
) -> None:
    router = APIRouter()

    @router.get(f"{prefix}/tenant_schema")
    async def get_tenant_schema() -> dict[str, Any]:
        return configs.tenant_schema.get_schema()

    @router.get(f"{prefix}/tenant_settings")
    async def get_tenant_settings(tenant: str):
        try:
            return await configs.self[tenant]
        except TenantNotFound as e:
            raise DoesNotExist(_("Tenant")) from e

    @router.post(f"{prefix}/tenant_settings")
    async def set_tenant_settings(setting: dict[str, Any], tenant: str):
        old_doc = (
            old_doc.model_dump()
            if (old_doc := await configs.tenant_schema.document.get(tenant))
            is not None
            else {}
        )
        try:
            doc = configs.tenant_schema.document.model_validate(
                old_doc | setting | {"id": tenant}
            )
            configs.tenant_schema.validate(doc)
            # ^check if its valid with current default configs
        except ValidationError as e:
            raise HTTPException(
                status_code=422,
                detail={
                    "error": _("Invalid tenant settings"),
                    "errors": e.errors(),
                },
            ) from e
        await doc.save()
        await configs.tenant_schema.cache.delete(tenant)
        # ^invalidate cache

    @router.get(f"{prefix}/reload")
    async def reload_endpoint() -> JSONResponse:
        reload_app()
        return JSONResponse(content={"status": "ok"})

    app.include_router(router, tags=["System"])



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from functools import lru_cache
from logging import Logger

import uvicorn
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from fastloom.launcher.settings import LauncherSettings
from fastloom.launcher.utils import (
    get_app,
    get_settings_cls,
    get_tenant_cls,
    is_installed,
)
from fastloom.logging.lifehooks import setup_logging
from fastloom.logging.settings import LoggingSettings
from fastloom.monitoring import InitMonitoring
from fastloom.observability.settings import ObservabilitySettings
from fastloom.settings.base import FastAPISettings
from fastloom.signals.depends import RabbitSubscriber, RabbitSubscriptable
from fastloom.signals.lifehooks import init_streams
from fastloom.signals.settings import RabbitmqSettings
from fastloom.tenant.settings import ConfigAlias as Configs

logger: Logger = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    service_app = get_app()
    await service_app.load()
    init_streams(service_app.stream_models)

    if Configs.cache_enabled:
        from aredis_om import Migrator

        await Migrator().run()

    async with service_app.lifespan_fn(app):
        yield


@lru_cache
def app():
    Configs(get_settings_cls(), get_tenant_cls())
    if isinstance(Configs[LoggingSettings].general, LoggingSettings):
        setup_logging(Configs[LoggingSettings].general)
    if isinstance(Configs[RabbitSubscriptable].general, RabbitmqSettings):
        RabbitSubscriber(Configs[RabbitSubscriptable].general)
    elif is_installed("aio-pika"):
        logging.warning("Settings Does Not Inherit from RabbitmqSettings")
    # ^IMPORTANT:rabbit has to init first
    service_app = get_app()
    with InitMonitoring(
        Configs[ObservabilitySettings].general,
        instruments=service_app.additional_instruments,
    ) as monitor:
        app = FastAPI(
            lifespan=lifespan,
            title=Configs[FastAPISettings].general.PROJECT_NAME,
            docs_url=f"{Configs[FastAPISettings].general.API_PREFIX}/docs",
            redoc_url=f"{Configs[FastAPISettings].general.API_PREFIX}/redoc",
            openapi_url=f"{Configs[FastAPISettings].general.API_PREFIX}/openapi.json",
            swagger_ui_oauth2_redirect_url=f"{Configs[FastAPISettings].general.API_PREFIX}/docs/oauth2-redirect",
            swagger_ui_init_oauth={
                "additionalQueryStringParams": {"browser": "false"},
            },
        )
        monitor.instrument(app)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    service_app.load_exception_handlers(app)
    service_app.load_healthchecks(app)
    service_app.load_system_endpoints(app)

    service_app.load_routes(app)
    service_app.load_mounts(app)
    if isinstance(Configs[RabbitSubscriptable].general, RabbitmqSettings):
        app.include_router(RabbitSubscriber.router)
    return app


def main():
    Configs(get_settings_cls(), get_tenant_cls())
    uvicorn.run(
        app=f"{__name__}:app",
        host="0.0.0.0",
        port=Configs[LauncherSettings].general.APP_PORT,
        reload=Configs[LauncherSettings].general.DEBUG,
        workers=Configs[LauncherSettings].general.WORKERS,
        factory=True,
    )



from contextlib import suppress
from typing import TYPE_CHECKING

from fastloom.cache.settings import RedisSettings
from fastloom.meta import SelfSustaining

_HAS_REDIS = False
with suppress(ImportError):
    from aredis_om import get_redis_connection
    from redis import Redis as SyncRedis
    from redis.asyncio import Redis
    from redis.exceptions import ConnectionError

    _HAS_REDIS = True

if TYPE_CHECKING:
    from redis import Redis as SyncRedis
    from redis.asyncio import Redis

if not TYPE_CHECKING and not _HAS_REDIS:
    SyncRedis = None
    Redis = None


class RedisHandler(SelfSustaining):
    enabled: bool = False
    redis: Redis
    sync_redis: SyncRedis

    def __init__(self, settings: RedisSettings) -> None:
        super().__init__()
        if not _HAS_REDIS:
            return
        self.redis = get_redis_connection(url=str(settings.REDIS_URL))
        self.sync_redis = SyncRedis.from_url(url=str(settings.REDIS_URL))
        with suppress(ConnectionError):
            self.enabled = self.sync_redis.ping()



from collections.abc import Callable, Coroutine
from functools import partial
from typing import Any

from faststream.rabbit.fastapi import RabbitRouter


class RabbitConnectionError(Exception): ...


async def check_rabbit_connection(router: RabbitRouter) -> None:
    try:
        await router.broker.ping(timeout=5)
    except Exception as er:
        raise RabbitConnectionError(f"RabbitMQ connection error: {er}") from er


def get_healthcheck(
    router: RabbitRouter,
) -> Callable[[], Coroutine[Any, Any, None]]:
    return partial(check_rabbit_connection, router=router)



from datetime import date, datetime, time
from zoneinfo import ZoneInfo

import jdatetime  # type: ignore[import-untyped]


def datetime_to_jalali(dt: datetime, date_only: bool = False) -> str:
    dt = dt.astimezone(ZoneInfo("Asia/Tehran"))
    return jdatetime.datetime.fromgregorian(datetime=dt).strftime(
        "%Y/%m/%d" if date_only else "%Y/%m/%d - %H:%M"
    )


def utcnow() -> datetime:
    return datetime.now(ZoneInfo("UTC"))


def datetime_to_timestamp(value: datetime) -> int:
    dt_utc = value.replace(tzinfo=ZoneInfo("UTC"))
    return int(dt_utc.timestamp())


def get_zero_time() -> datetime:
    return datetime.combine(date.today(), time())



from pydantic import (
    BaseModel,
    Field,
    computed_field,
)

from fastloom.meta import infer_project_name


class ProjectSettings(BaseModel):
    PROJECT_NAME: str = Field(default_factory=infer_project_name)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def API_PREFIX(self) -> str:
        return f"/api/{self.PROJECT_NAME}"


class FastAPISettings(ProjectSettings):
    DEBUG: bool = True


class MonitoringSettings(ProjectSettings):
    ENVIRONMENT: str



import functools
from collections.abc import Awaitable, Callable

from pymongo import AsyncMongoClient
from pymongo.asynchronous.client_session import AsyncClientSession

from fastloom.db.lifehooks import get_mongo_client


class MongoTransactionManager:
    def __init__(self, mongo_uri: str):
        self.mongo_uri = mongo_uri

    async def __aenter__(self) -> AsyncClientSession:
        self.client: AsyncMongoClient = await get_mongo_client(self.mongo_uri)
        self.session: AsyncClientSession = self.client.start_session()
        await self.session.start_transaction()
        return self.session

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: object | None = None,
    ):
        if exc_type:
            await self.session.abort_transaction()
        else:
            await self.session.commit_transaction()
        await self.session.end_session()
        await self.client.close()


def with_transaction[T, **P](
    mongo_uri: str,
) -> Callable[[Callable[..., Awaitable[T]]], Callable[..., Awaitable[T]]]:
    def transaction_wrapper(
        func: Callable[P, Awaitable[T]],
    ) -> Callable[P, Awaitable[T]]:
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with MongoTransactionManager(mongo_uri) as session:
                kwargs["session"] = session
                return await func(*args, **kwargs)

        return wrapper

    return transaction_wrapper



from typing import TYPE_CHECKING, Any

from faststream.opentelemetry.middleware import TelemetryMiddleware
from faststream.rabbit.opentelemetry.provider import (
    RabbitTelemetrySettingsProvider,
)
from opentelemetry.metrics import Meter, MeterProvider
from opentelemetry.trace import TracerProvider
from pydantic import BaseModel

if TYPE_CHECKING:
    from aio_pika import IncomingMessage
    from faststream.message import StreamMessage
    from faststream.rabbit.response import RabbitPublishCommand


def message_body_to_str(message_body) -> str:
    if isinstance(message_body, BaseModel):
        return message_body.model_dump_json()
    return str(message_body)


class RabbitPayloadTelemetrySettingsProvider(RabbitTelemetrySettingsProvider):
    def get_publish_attrs_from_cmd(
        self,
        kwargs: "RabbitPublishCommand",
    ) -> dict[str, Any]:
        ret = super().get_publish_attrs_from_cmd(kwargs)
        try:
            message_body_str = message_body_to_str(kwargs.body)
        except ValueError:
            return ret
        return ret | {"messaging.message.body": message_body_str}

    def get_consume_attrs_from_message(
        self, msg: "StreamMessage[IncomingMessage]"
    ) -> dict[str, Any]:
        return super().get_consume_attrs_from_message(msg) | {
            "messaging.message.body": msg.body.decode()
        }


class RabbitPayloadTelemetryMiddleware(TelemetryMiddleware):
    def __init__(
        self,
        *,
        tracer_provider: TracerProvider | None = None,
        meter_provider: MeterProvider | None = None,
        meter: Meter | None = None,
        include_messages_counters: bool = False,
    ) -> None:
        super().__init__(
            settings_provider_factory=(
                lambda _: RabbitPayloadTelemetrySettingsProvider()  # type: ignore
            ),
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
            meter=meter,
            include_messages_counters=include_messages_counters,
        )



from collections.abc import Callable, Coroutine
from typing import Annotated, Any

import httpx
from fastapi import Depends, HTTPException, Request
from fastapi.security import OAuth2, OpenIdConnect
from jose.jwt import get_unverified_claims

from fastloom.auth import Claims
from fastloom.auth.schemas import UserClaims
from fastloom.auth.settings import IAMSettings, IntrospectionResponse


class OptionalJWTAuth:
    settings: IAMSettings
    _security_scheme: OAuth2 | OpenIdConnect | None = None

    def __init__(self, settings: IAMSettings):
        self.settings = settings

        if self.settings.oidc_enabled:
            assert self.settings.OIDC_URL is not None
            self._security_scheme = OpenIdConnect(
                openIdConnectUrl=self.settings.OIDC_URL,
                scheme_name="OIDC",
                auto_error=False,
            )
        elif self.settings.oauth2_enabled:
            self._security_scheme = OAuth2(
                flows=self.settings.flows, auto_error=False
            )

    async def _introspect(
        self, token: Annotated[str, Depends(_security_scheme)]
    ):
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.post(
                f"{self.settings.IAM_SIDECAR_URL}/introspect",
                json=dict(token=token),
            )
        if response.status_code != 200:
            raise HTTPException(status_code=403, detail=response.text)
        data = IntrospectionResponse.model_validate(response.json())
        if not data.active:
            raise HTTPException(status_code=403, detail="Inactive token")

    def _transform_bearer(self, token: str) -> str:
        if token.startswith("Bearer "):
            return token.removeprefix("Bearer ").strip()
        return token

    async def _acl(
        self,
        request: Request,
        token: Annotated[str, Depends(_security_scheme)],
    ) -> None:
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.post(
                url=f"{self.settings.IAM_SIDECAR_URL}/acl",
                json={
                    "token": token,
                    "endpoint": request.url.path,
                    "method": request.method,
                },
            )

        if response.status_code != 200:
            raise HTTPException(status_code=403, detail=response.text)
        if not response.json():
            raise HTTPException(status_code=403)

    @classmethod
    def _parse_token(cls, token: str) -> UserClaims:
        return UserClaims.model_validate(get_unverified_claims(token))

    async def _validate_token(
        self, token: str, request: Request
    ) -> UserClaims:
        token = self._transform_bearer(token)
        if self.settings.INTROSPECT:
            await self._introspect(token)
        if self.settings.ACL:
            await self._acl(request, token)
        claims = self._parse_token(token)
        Claims.set(claims)
        return claims

    @property
    def get_token(
        self,
    ) -> Callable[..., Coroutine[Any, Any, str | None]]:
        async def _inner(
            token: Annotated[str | None, Depends(self._security_scheme)],
        ) -> str | None:
            if token is None:
                return None
            return self._transform_bearer(token)

        return _inner

    @property
    def get_claims(
        self,
    ) -> Callable[..., Coroutine[Any, Any, UserClaims | None]]:
        async def _inner(
            request: Request,
            token: Annotated[str | None, Depends(self._security_scheme)],
        ) -> UserClaims | None:
            if token is None:
                return None

            return await self._validate_token(token, request)

        return _inner


class JWTAuth(OptionalJWTAuth):
    def __init__(self, settings: IAMSettings):
        super().__init__(settings)
        if self._security_scheme is not None:
            self._security_scheme.auto_error = True

    @property
    def get_claims(self) -> Callable[..., Coroutine[Any, Any, UserClaims]]:
        async def _inner(
            request: Request,
            token: Annotated[str, Depends(self._security_scheme)],
        ) -> UserClaims:
            return await self._validate_token(token, request)

        return _inner

    @property
    def get_token(
        self,
    ) -> Callable[..., Coroutine[Any, Any, str]]:
        async def _inner(
            token: Annotated[str, Depends(self._security_scheme)],
        ) -> str:
            return self._transform_bearer(token)

        return _inner
